
// version finale
/*
Image.cpp - Library for due

Image.h - Library for due

by Cordier Yves Jan 2015 Version initiale du 01/02/2015

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

Cette librairie permet la manipilation d'images en memoire
ainsi que la connection d'une camera via l'objet camera
Elle permet aussi la reconaissance de pièces sur et d'avoir 
les caractéristiques géométriques de la pièce photographiée

Historique des versions
- V1.0 01/06/2016
- V1.01 21/06/2016
	Modification amelioration de la méthode  UnStain() (plus rapide) et qui fonctionne en multichrome (256 couleurs)
	Creation de la méthode UnStainTol(tol) qui permet de mettre une 
	 tolerance par rapport au N° de gris affiché (+- tolerance)
	 Tol=0 signifie exactitude soit elimination des taches de couleur isolées quelque soit la couleur
	Ajout d'une procedure de swap de deux images SwapImage(I1,I2) by Cordier Yves Jan 2015 Version du 01/02/2015

*/

/******************************************************************************
* Includes
******************************************************************************/
#include "Image.h" // Attention aux majuscules/minuscules dans le nom du fichier
#include <math.h>


// *************
// L'objet Image
// *************

Image::Image() // Le constructeur vide
{ _Tx = 0;
	_Ty = 0;
	_content = NULL; // variable vide

}

Image::Image(unsigned int Tx, unsigned int Ty) // Le constructeur amélioré
{ _content = (byte*) calloc (Tx * Ty, sizeof(byte)); // Creation d'une zone mémoire remplie de 0
	if (_content == NULL) // si plus de memoire dispo
	{
		_Tx = 0;
		_Ty = 0;
	}
	else
	{ _Tx = Tx;
		_Ty = Ty;
	}
}

Image::~Image() // le destructeur
{
	free(_content); // libère la memoire
	_Tx = 0; _Ty = 0;
};

// Lecture d'un pixel
int Image::GetPixel(int x, int y) // lecture de la couleur d'un pixel
{
	if ((x >= 0) && (x < _Tx) && (y >= 0) && (y < _Ty))
	{
		return byte( *(_content + x + y * _Tx));
	}
	else
	{
		return -1; // out of image zone
	}
}

// Ecriture d'un pixel
void Image::PutPixel(unsigned int x, unsigned int y, byte c) // ecriture d'un pixel
{
	if ((x >= 0) && (x < _Tx) && (y >= 0) && (y < _Ty))
	{
		*(_content + x + y * _Tx) = c;
	}
}


void Image::Bichromie(byte seuil) // passage en bichromie (noir et blanc)
{
	for (unsigned int i = 0; i < (_Tx * _Ty); i++)
	{ if (byte(*(_content + i)) > seuil)
		{
			*(_content + i) = 255;
		}
		else
		{
			*(_content + i) = 0;
		}
	}
}
void Image::Negatif() // mise en negatif
{
	for (unsigned int i = 0; i <( _Tx * _Ty); i++)
	{
		byte c=*(_content + i); 
		*(_content + i)=~c;// Not bit à bit
	} 
}

void Image::Flipx() // flip de l'image / axe vertical
{ for (unsigned int j = 0; j < _Ty; j++)
	{
		for (unsigned int i = 0; i < (_Tx / 2); i++)
		{
			byte sto = *(_content + j * _Tx + i);
			*(_content + j * _Tx + i) = *(_content + j * _Tx + (_Tx - i));
			*(_content + j * _Tx + (_Tx - i)) = sto;
		}
	}
}
void Image::Flipy() // flip de l'image / axe horizontal
{ for (unsigned int i = 0; i < _Tx; i++)
	{
		for (unsigned int j = 0; j < (_Ty / 2); j++)
		{
			byte sto = *(_content + j * _Tx + i);
			*(_content + j * _Tx + i) = *(_content + (_Ty - j) * _Tx + i);
			*(_content + (_Ty - j)*_Tx + i) = sto;
		}
	}
}
unsigned int Image::SiseX() // Taille de l'image sur x en pixel
{
	return _Tx;
}
unsigned int Image::SiseY() // Taille de l'image sur y en pixel
{
	return _Ty;
}

void Image::Line(int x0, int y0, int x1, int y1, int color) // tracé d'une ligne
{
	int dx = abs(x1 - x0);
	int sx = x0 < x1 ? 1 : -1; // int sx; if (x0<x1) {sx=1;} else {sx=-1};
	int dy = abs(y1 - y0);
	int sy = y0 < y1 ? 1 : -1; // int sy; if (y0<y1) {sy=1;} else {sy=-1}
	int err = (dx > dy ? dx : -dy) / 2;
	int e2;

	for (;;)
	{
		if (color < 0)
		{
			PutPixel(x0, y0, ~GetPixel(x0, y0)); // en opposition video
		}
		else
		{
			PutPixel(x0, y0, byte(color)); // avec la couleur donnée
		}

		if (x0 == x1 && y0 == y1) break;
		e2 = err;
		if (e2 > -dx) {
			err -= dy;
			x0 += sx;
		}
		if (e2 < dy) {
			err += dx;
			y0 += sy;
		}
	}
}

void Image::Circle(int x1, int y1, int radius, int couleur) // tracé d'un cercle
{
	// inspiré de http://www.ecse.rpi.edu/Homepages/wrf/Research/Short_Notes/bresenham.html
	int d, x, y, x2m1, max;
	y = radius;
	d = -radius;
	x2m1 = -1;
	max = (int)(radius / sqrt(2.0));
	for (x = 0; x <= max; x++)
	{
		x2m1 += 2;
		d += x2m1;
		if (d >= 0)
		{
			y--;
			d -= (y << 1);
		}
		if (couleur > 0)
		{
			PutPixel(x1 + x, y1 + y, couleur);
			PutPixel(x1 - x, y1 + y, couleur);
			PutPixel(x1 + x, y1 - y, couleur);
			PutPixel(x1 - x, y1 - y, couleur);
			PutPixel(x1 + y, y1 + x, couleur);
			PutPixel(x1 - y, y1 + x, couleur);
			PutPixel(x1 + y, y1 - x, couleur);
			PutPixel(x1 - y, y1 - x, couleur);
		}
		else
		{

			PutPixel(x1 + x, y1 + y, ~GetPixel(x1 + x, y1 + y));
			PutPixel(x1 - x, y1 + y, ~GetPixel(x1 - x, y1 + y));
			PutPixel(x1 + x, y1 - y, ~GetPixel(x1 + x, y1 - y));
			PutPixel(x1 - x, y1 - y, ~GetPixel(x1 - x, y1 - y));
			PutPixel(x1 + y, y1 + x, ~GetPixel(x1 + y, y1 + x));
			PutPixel(x1 - y, y1 + x, ~GetPixel(x1 - y, y1 + x));
			PutPixel(x1 + y, y1 - x, ~GetPixel(x1 + y, y1 - x));
			PutPixel(x1 - y, y1 - x, ~GetPixel(x1 - y, y1 - x));
		}
	}
}

void Image::Disc(int x1, int y1, int radius, int couleur) // tracé d'un disque
{
	int x;
	int y;
	int ysto = 10000;
	radius = abs(radius);
	int maxi = round(radius / sqrt(2));
	for (y = maxi + 1; y <= radius; y++)
	{
		x = int(sqrt(sq(radius) - sq(y)));
		if (y != ysto)
		{
			HorizLine(x1 + x, y1 + y, x1 - x, couleur);
			HorizLine(x1 + x, y1 - y, x1 - x, couleur);
			VertLine(x1 + y, y1 + x, y1 - x, couleur);
			VertLine(x1 - y, y1 + x, y1 - x, couleur);
			ysto = y;
		}
	}
	Rect(x1 - maxi, y1 - maxi, x1 + maxi, y1 + maxi, couleur);
}

void Image::HorizLine( int x1, int y1, int x2, int couleur) // ligne horizontale
{
	int i;
	if (x2<x1) {i=x1;x1=x2;x2=i;} // inversion
	for (i = x1; i <= x2; i++)
	{ if (couleur > 0)
		{
			PutPixel(i, y1, couleur);
		}
		else
		{
			PutPixel(i, y1, ~GetPixel(i, y1));
		}
	}
}

void Image::VertLine(int x1, int y1, int y2, int couleur) // ligne verticale
{
	int i;
	if (y2<y1) {i=y1;y1=y2;y2=i;} // inversion
	for (i = y1; i <= y2; i++)
	{ if (couleur > 0)
		{
			PutPixel(x1, i, couleur);
		}
		else
		{
			PutPixel(x1, i, ~GetPixel(x1, i));
		}
	}
}

void Image::Box(int x1, int y1, int x2, int y2, int couleur) // Tracé d'une boite vide
{
	HorizLine( x1, y1, x2, couleur);
	HorizLine( x1, y2, x2, couleur);
	VertLine( x1, y1, y2, couleur);
	VertLine( x2, y1, y2, couleur);
	if (couleur < 0) // on corrige les effets d'angles
	{ PutPixel(x1, y1, couleur);
		PutPixel(x1, y2, couleur);
		PutPixel(x2, y2, couleur);
		PutPixel(x2, y1, couleur);
	}
}

void Image::Rect(int x1, int y1, int x2, int y2, int couleur) // Tracé et remplissage d'un rectangle
{
	int i, j;
	if (x2 < x1) {
		i = x2;
		x2 = x1;
		x1 = i;
	};
	if (y2 < y1) {
		i = y2;
		y2 = y1;
		y1 = i;
	};
	for (i = x1; i <= x2; i++)
	for (j = y1; j <= y2; j++)
	PutPixel(i, j, couleur);
}

void Image::Fill(byte fond)
{for (unsigned int i=0; i<_Tx*_Ty;i++)
	{ *(_content+i)=fond;}
}
void* Image::Contenu()
{
	return _content;
}
void Image::Rotate(int Cx, int Cy, float a,int fond)
{ 
	byte * buffer;;
	buffer = (byte*) malloc (_Tx*_Ty);
	float Ct=cos(a*M_PI/180);
	float St=sin(a*M_PI/180);
	for (unsigned int i=0; i<_Tx*_Ty; i++) {buffer[i]=fond;}; // vide le buffer avec couleur fond
	for (unsigned int i=0; i<_Tx; i++)
	{ for (unsigned int j=0; j<_Ty;j++)
		{
			int nx=int(round(Cx+Ct*(i-Cx)-St*(j-Cy)));
			int ny=int(round(Cy+St*(i-Cx)+Ct*(j-Cy)));
			
			if ((nx>=0)&(nx<int(_Tx))&(ny>=0)&(ny<int(_Ty)))
			{buffer[nx+(ny)*_Tx]=GetPixel(i,j);};
		}
	}
	for (unsigned int i=0; i<_Tx*_Ty;i++) {*(_content+i)=buffer[i];} // recopie de l'image transformée
	free (buffer);
	
}

void Image::Translate(int dx, int  dy, byte fond)
{
	byte * buffer;;
	buffer = (byte*) malloc (_Tx*_Ty);
	for (unsigned int i=0; i<_Tx*_Ty; i++) {buffer[i]=fond;}; // vide le buffer avec couleur fond
	for (unsigned int i=0; i<_Tx; i++)
	{ for (unsigned int j=0; j<_Ty;j++)
		{
			int nx=i+dx;
			int ny=j+dy;
			if ((nx>=0)&(nx<int(_Tx))&(ny>=0)&(ny<int(_Ty)))
			{buffer[nx+(ny)*_Tx]=GetPixel(i,j);};
		}
	}
	for (unsigned int i=0; i<_Tx*_Ty;i++) {*(_content+i)=buffer[i];} // recopie de l'image transformée
	free (buffer);
	
}


void Image::Stain(unsigned int n)
{
	randomSeed(millis()); 
	for (unsigned int i=1; i<=n;i++)
	{PutPixel(random(_Tx),random(_Ty),random(2)*255);}
}
void Image::UnStain()
{
	UnStainTol(0);
}

void Image::UnStainTol(byte tolerance)
{
	int c; // la couleur du point a traiter
	int c1; // la couleur du point à coté
	// c et c1 sont des int pour pouvoir calculer c-c1 et abs(c-c1)
	
	long col=0; // la somme des couleur des points environnant
	byte n=0; // le nombre de points différent environnant
	byte ident; // le nombre de point équivalent environnant
	for (unsigned int i=0; i<_Tx; i++) // balayage sur les x
	{
		for (unsigned int j=0; j<_Ty; j++) // balayage sur les y
		{
			n=0;
			ident=0;
			col=0;
			c=GetPixel(i,j);
			// traitement des points environnants
			if (i>0) // à gauche
			{
				c1=GetPixel(i-1,j);
				if ((abs(c1-c)>tolerance)) {n++; col+=c1;} else {ident++;}
			}
			if (i<_Tx-1)  // a droite
			{
				c1=GetPixel(i+1,j);
				if ((abs(c1-c)>tolerance)) {n++; col+=c1;}else {ident++;}
			}
			if (j>0) // au dessus
			{
				c1=GetPixel(i,j-1);
				if ((abs(c1-c)>tolerance)) {n++; col+=c1;}else {ident++;}
			}
			if (j<_Ty-1) // au dessous
			{
				c1=GetPixel(i,j+1);
				if ((abs(c1-c)>tolerance)) {n++; col+=c1;}else {ident++;}
			}
			if ((i>0) & (j>0)) // au dessus à gauche
			{
				c1=GetPixel(i-1,j-1);
				if ((abs(c1-c)>tolerance)) {n++; col+=c1;}else {ident++;}
			}
			if ((i<_Tx-1) & (j<_Ty-1)) // au dessous à droite
			{
				c1=GetPixel(i+1,j+1);
				if ((abs(c1-c)>tolerance)) {n++; col+=c1;}else {ident++;}
			}
			if ((i>0)& (j<_Ty-1))  // au dessous à gauche
			{
				c1=GetPixel(i-1,j+1);
				if ((abs(c1-c)>tolerance)) {n++; col+=c1;}else {ident++;}
			}
			if ((i<_Tx-1)&(j>0))  // au dessus à droite
			{
				c1=GetPixel(i+1,j-1);
				if ((abs(c1-c)>tolerance)) {n++; col+=c1;}else {ident++;}
			}
			// traitement final en fonction de l'analyse
			if (ident==0) {PutPixel(i,j,col/n);} // pas de point equivalen aux alentours => on affecte la valeur moyenne au point
		}
	}
}




void Image::Cdg_inertie_seuil(float &surf, float &Cx, float &Cy,float &Ix, float &Iy, float &Ixy, int seuil)// caractéristiques de l'image (pts au dessus du seuil);
{
	
	float resultats=0;
	float resultatgx=0;
	float resultatgy=0;
	float resultatix=0;
	float resultatiy=0;
	float resultatixy=0;
	for (unsigned int i=0; i<_Tx-1; i++)
	{
		for (unsigned int j=0; j<_Ty-1; j++)
		{
			if (GetPixel(i,j)>=seuil) 
			{
				resultats+=1;
				resultatgx+=i;
				resultatgy+=j;
				resultatix+=j*j;
				resultatiy+=i*i;
				resultatixy+=i*j;
			}

		}
		
	}
	surf=resultats;
	if (surf!=0) 
	{
		Cx=resultatgx/surf;
		Cy=resultatgy/surf;
	}
	else
	{
		if (_Tx>0) {Cx=(_Tx-1)/2;} else {Cx=0;}
		if (_Ty>0) {Cy=(_Ty-1)/2;} else {Cy=0;}
	}
	Ix=resultatix;
	Iy=resultatiy;
	Ixy=resultatixy;
	
	

}


void Image::Calcul_Caracteristiques()
{ // On ramene la matrice au cdg
	float Ix = _Ixx - _Surf * sq(_Cdgy);
	float Iy = _Iyy - _Surf * sq(_Cdgx);
	float Ixy = _Ixy - _Surf * _Cdgx * _Cdgy;

	// Determination de l'angle de rotation de la pièdce / repere de l'image
	// Calcul de l'angle du repere principal / repere image
	float Di=Ix - Iy;
	_Angle =(atan2(-2 * Ixy, Di) / 2)*180/3.1415962; // angle en degré

	// On tourne la matrice de _angle pour obtenir les bonnes valeurs de _Ixcdg et _Iycdg
	float A=(Ix+Iy)/2;
	float B=sqrt(sq(Di/2)+sq(Ixy));
	_Ixcdg = A+B;
	_Iycdg = A-B;
	
};

void Image::Genere_aleatoire()
{
	int a=random(180)-90;
	int t=random(3); // pièce de type 0,1,2
	int l=_Tx/5;
	Fill(0); // met l'image à 'blanc' (vide)
	switch (t)
	{case 0 : // pièce rectangulaire 
		Rect(_Tx/2-l,_Ty/2-l/2,_Tx/2+l,_Ty/2+l/2,255);
		break;
	case 1 : // pièce circulaire
		Disc(_Tx/2,_Ty/2,l,255);
		break;
	case 2 : //pièce trefle
		Disc(_Tx/2,_Ty/2+l/2,l*2/3,255);
		Disc(_Tx/2-l/3,_Ty/2-l/4,l*2/3,255);
		Disc(_Tx/2+l/3,_Ty/2-l/4,l*2/3,255);
		Rect(_Tx/2-l/5,_Ty/2,_Tx/2+l/5,_Ty/2-l,255);
		break;
	}
	// rotation autour du centre
	Rotate(_Tx/2,_Ty/2,a,0);
	// decalage aleatoire
	Translate(random(2*l)-l,random(2*l)-l,0);

}

// *********************
// Partie à implementer
// *********************

void Image::Calcul_Ixx()
{
	float x=0;
	for (unsigned int i = 0; i < _Tx; i++)
	{
		for (unsigned int j = 0; j < _Ty; j++)
		{
			if (boolean(GetPixel(i,j)))
			{
				x+=j*j;
				
			}
		}
	}
	_Ixx=x;
}
void Image::Calcul_Iyy()
{
	float x=0;
	for (unsigned int i = 0; i < _Tx; i++)
	{
		for (unsigned int j = 0; j < _Ty; j++)
		{
			if (boolean(GetPixel(i,j)))
			{
				x+=i*i;
				
			}
		}
	}
	_Iyy=x;
}

void Image::Calcul_Ixy() // Méthode Calcul_Ixy de la classe Image
{float xy=0; // on vide la valeur courante
	for (unsigned int i = 0; i < _Tx; i++) // boucle sur les abcisses
	{
		for (unsigned int j = 0; j < _Ty; j++) // boucle sur les ordonnées
		{
			if (boolean(GetPixel(i,j))) // si couleur pixel >0
			{
				xy+=i*j; //  prise en compte du pixel
			}
		} // fin boucle ordonnées
	} // fin boucle abcisses
	
	_Ixy=xy; // on affecte l'attribut de Image
}

void Image::Calcul_Surf()
{unsigned int a=0;
	for (unsigned int i = 0; i < _Tx; i++)
	{
		for (unsigned int j = 0; j < _Ty; j++)
		{
			if (boolean(GetPixel(i,j)))
			{
				a++;
			}
		}
	}
	_Surf=a;
}

void Image::Calcul_Cdgx()
{float x=0;
	unsigned int a=0;
	for (unsigned int i = 0; i < _Tx; i++)
	{
		for (unsigned int j = 0; j < _Ty; j++)
		{
			if (boolean(GetPixel(i,j)))
			{
				x=x+i;
				a++;
			}
		}
	}
	x=x/a;
	_Cdgx=x;
}

void Image::Calcul_Cdgy()
{float y=0;
	unsigned int a=0;
	for (unsigned int i = 0; i < _Tx; i++)
	{
		for (unsigned int j = 0; j < _Ty; j++)
		{
			if (boolean(GetPixel(i,j)))
			{
				y=y+j;
				a++;
			}
		}
	}
	y=y/a;
	_Cdgy=y;
}

void Image::Tout_calculs_seuil(int seuil)
{
	Cdg_inertie_seuil(_Surf,_Cdgx,_Cdgy,_Ixx,_Iyy,_Ixy, seuil);// Caractéristiques inertielles en monochrome au dessus du seuil;
	Calcul_Caracteristiques();
}

void SwapImage(Image &I1, Image &I2)
{
	unsigned int Sto;
	// Les tailles à inverser
	Sto=I1._Tx;I1._Tx=I2._Tx; I2._Tx=Sto;
	Sto=I1._Ty;I1._Ty=I2._Ty; I2._Ty=Sto;
	float Fsto;
	// Les attributs de calcul de caractéristiques d'image à inverser
	Fsto=I1._Iyy;I1._Iyy=I2._Iyy; I2._Iyy=Fsto;
	Fsto=I1._Ixx;I1._Ixx=I2._Ixx; I2._Ixx=Fsto;
	Fsto=I1._Ixy;I1._Ixy=I2._Ixy; I2._Ixy=Fsto;
	Fsto=I1._Ixcdg;I1._Ixcdg=I2._Ixcdg; I2._Ixcdg=Fsto;
	Fsto=I1._Iycdg;I1._Iycdg=I2._Iycdg; I2._Iycdg=Fsto;
	Fsto=I1._Cdgx;I1._Cdgx=I2._Cdgx; I2._Cdgx=Fsto;
	Fsto=I1._Cdgy;I1._Cdgy=I2._Cdgy; I2._Cdgy=Fsto;
	Fsto=I1._Angle;I1._Angle=I2._Angle; I2._Angle=Fsto;
	Fsto=I1._Surf;I1._Surf=I2._Surf; I2._Surf=Fsto;
	byte* StoPointer; // Un pointeur sur un pseudo byte
	// Inversion des contenus des images
    StoPointer=I1._content; I1._content=I2._content; I2._content=StoPointer;
}

