/*
Image.h - Library for due

by Cordier Yves Jan 2015 Version initiale du 01/02/2015

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

Cette librairie permet la manipilation d'images en memoire
ainsi que la connection d'une camera via l'objet camera
Elle permet aussi la reconaissance de pièces sur et d'avoir 
les caractéristiques géométriques de la pièce photographiée

Historique des versions
- V1.0 01/06/2016
- V1.01 21/06/2016
	Modification amelioration de la méthode  UnStain() (plus rapide) et qui fonctionne en multichrome (256 couleurs)
	Creation de la méthode UnStainTol(tol) qui permet de mettre une 
	 tolerance par rapport au N° de gris affiché (+- tolerance)
	 Tol=0 signifie exactitude soit elimination des taches de couleur isolées quelque soit la couleur
	Ajout d'une procedure de swap de deux images SwapImage(I1,I2) 
*/

#ifndef image_h
#define image_h
#if defined(ARDUINO) && ARDUINO >= 100
#include "Arduino.h"
#else
#include "WProgram.h"
#endif
#include <HardwareSerial.h>
// #ifdef _VARIANT_ARDUINO_DUE_X_class
// Zone de données
#define _Epsilon 0.00001
#define _pi 3.1415962


class Image
{
	// private:
public: // ces attributs seront visibles dans les classes derivés mais non dans le pgm principal
	unsigned int _Tx = 0;
	unsigned int _Ty = 0;
	// Les attributs de calcul de caractéristiques d'image
	float _Ixx = 0;
	float _Iyy = 0;
	float _Ixy = 0;
	float _Ixcdg = 0;
	float _Iycdg = 0;
	float _Cdgx = 0;
	float _Cdgy = 0;
	float _Angle = 0;
	float _Surf = 0;
	
	byte* _content; // Un pointeur sur un pseudo byte
	// en fait une table de byte Chaque pixel est représenté par un byte dans une structure linéaire ligne à ligne

	// Les constructeurs
	Image(); // Le constructeur par defaut
	Image(unsigned int Tx, unsigned int Ty); // Le constructeur amélioré
	// Le destructeur
	~Image();
	// Les procedures et fonctions d'acquisition-transfert-visualisation

	// Les procedures et fonctions de traitement d'image
	int GetPixel(int x,  int y); // donne la couleur du pixel (x,y) -1 si hors zone
	void PutPixel(unsigned int x, unsigned int y, byte c); // ecrit le pixel (x,y) à la couleur c
	void Line(int x0, int y0, int x1, int y1, int color);  // trace une ligne de (x0,y0) à (x1,y1) avec la couleur c si >0 en negatif si <0
	// Nota : Les points P0 et P1 peuvent être en dehors de la fenètre de tracé
	// Suivant l'algorithme de Bresenham
	void HorizLine( int x1, int y1, int x2, int couleur); // ligne horizontale (avec le même code de couleur)
	void VertLine(int x1, int y1, int y2, int couleur); // lignes verticales (avec le même code de couleur)
	void Box(int x1, int y1, int x2, int y2, int couleur); // une boite vide (avec le même code de couleur)
	void Rect(int x1, int y1, int x2, int y2, int couleur); // des rectangles plein (avec le même code de couleur)
	void Circle(int x1, int y1, int radius, int couleur); // des cercles (avec le même code de couleur)
	void Disc(int x1, int y1, int radius, int couleur); // des disques (avec le même code de couleur)
	void Bichromie(byte seuil); // converi l'image en bichromie
	void Negatif(); // passe l'image en negatif
	void Flipx(); // retourne l'image autour de l'axe y
	void Flipy(); // retourne l'image autour de l'axe x
	void Rotate(int Cx, int Cy, float a,int color_fond);
	void Translate(int dx, int  dy, byte fond);
	void Fill(byte fond);
	unsigned int SiseX(); // renvoie la taille de l'image en nb de pixel sur x  0 .. (Tx-1)
	unsigned int SiseY(); // renvoi la taille de l'image en nb de pixel sur y   0 .. (Ty-1)
	void Stain(unsigned int n); // génère des taches alleatoires
	void UnStain(); // Detache lae taches 1 point en monochrome
	void UnStainTol(byte tolerance);// Detache les taches 1 point avec une tolerance de couleur tol
	// Les procedures de calcul des caractéristiques géométriques
	void Cdg_inertie_seuil(float &surf, float &Cx, float &Cy,float &Ix, float &Iy, float &Ixy, int seuil);// caractéristiques du cdg des pts au dessus du seuil;
	// Les methodes à implementer
	void Calcul_Ixx();
	void Calcul_Iyy();
	void Calcul_Ixy();
	void Calcul_Surf();
	void Calcul_Cdgx();
	void Calcul_Cdgy();
	
	//
	// La méthode de calcul générale pour trouver les inerties propres
	void Calcul_Caracteristiques();
	void Tout_calculs_seuil(int seuil); // effectue dans le bon ordre tous les calculs sur l'image
	// Pour acceder au contenu
	void* Contenu();
	void Genere_aleatoire();
};

void SwapImage(Image &I1, Image &I2);
#endif